"""
System Event queue is a background thread that runs several repeated tasks in a single thread
"""
import logging
import datetime
import threading
import time

from sqlalchemy.exc import SQLAlchemyError

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)

_SYSTEM_EVENTS = []

_debug = False
_run_interval = None
_min_interval = None
_configured = False


def init(app):
    global _debug, _run_interval, _min_interval, _configured

    if _configured:
        raise Exception('Tried in initialise system events twice!')

    _debug = app.config.get('SYSTEM_EVENT_DEBUG', False)
    _run_interval = app.config['SYSTEM_EVENT_RUN_INTERVAL']
    _min_interval = app.config['SYSTEM_EVENT_MIN_INTERVAL']

    system_event_thread = _SystemEventThread()
    system_event_thread.start()


def queue_event(event):
    log.info('Queueing System Event: %s' % event)
    log.debug('Event will run at %s and then every %s seconds' % (event.next_run, event.run_every))
    _SYSTEM_EVENTS.append(event)


class SystemEvent(object):
    def __init__(self, first_run=None, run_every=None):
        """
        :param first_run: datetime - when this event should first run
        :param run_every: number of seconds between repeats of this task or None for no repeat
        """
        if first_run is None:
            self.next_run = datetime.datetime.utcnow()
        else:
            self.next_run = first_run

        self.run_every = run_every

        self.error = None
        self.done = False

    def run_job(self):
        from models import db

        try:
            if _debug:
                log.info('Running system event: %s' % self)
            self.process()
        except SQLAlchemyError as e:
            log.exception('Database error in system event %s - rolling back to try to recover' % self)
            db.session.rollback()
            self.error = str(e)
        except Exception as e:
            log.exception('Error running system event: %s' % self)
            self.error = str(e)

        if self.run_every:
            interval = datetime.timedelta(seconds=self.run_every)
            self.next_run = datetime.datetime.utcnow() + interval
            if _debug:
                log.info('Next run in %s at %s' % (interval, self.next_run))
        else:
            self.done = True

    def process(self):
        raise NotImplementedError()


class _SystemEventThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.daemon = True

    def run(self):
        from models import db
        global _SYSTEM_EVENTS

        log.info('System event thread starting up')
        log.info('Checking for system events every %s seconds' % _run_interval)

        # Start the loop
        while True:
            # Default to run_interval in case we crash
            sleep_time = _run_interval

            try:
                if _debug:
                    log.info('System Event Queue - waking up')
                now = datetime.datetime.utcnow()

                # Load all of the pending cron jobs
                for job in _SYSTEM_EVENTS:
                    if job.next_run < now:
                        job.run_job()

                _SYSTEM_EVENTS = [job for job in _SYSTEM_EVENTS if not job.done]

                # Now that we've run all jobs, calculate the next run time
                dt = datetime.datetime.utcnow() - now
                if _debug:
                    log.info('System event cycle took %s' % dt)
                sleep_time -= dt.seconds
                if sleep_time < _min_interval:
                    log.warning('System event cycle took too long!  Took %s when desired interval is %s seconds'
                                % (dt, _run_interval))
                    sleep_time = _min_interval

            except SQLAlchemyError:
                log.exception('Database error in System event thread - rolling back to try to recover')
                db.session.rollback()

            except Exception:
                log.exception('Error in System event thread')

            if _debug:
                log.info('Sleeping for %s seconds' % sleep_time)
            time.sleep(sleep_time)

        log.info('System event thread shutting down')



"""
Image processing utility functions
"""

import logging

import PIL.Image
import PIL.ImageEnhance

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


def resize_image_to_fit(image, dest_w, dest_h):
    """
    Resize the image to fit inside dest rectangle.  Resultant image may be smaller than target
    :param image: PIL.Image
    :param dest_w: Target width
    :param dest_h: Target height
    :return: Scaled image
    """
    dest_w = float(dest_w)
    dest_h = float(dest_h)
    dest_ratio = dest_w / dest_h

    # Calculate the apect ratio of the image
    src_w = float(image.size[0])
    src_h = float(image.size[1])
    src_ratio = src_w / src_h

    if src_ratio < dest_ratio:
        # Image is tall and thin - we need to scale to the right height and then pad
        scale = dest_h / src_h
        scaled_h = dest_h
        scaled_w = src_w * scale

    else:
        # Image is short and wide - we need to scale to the right height and then crop
        scale = dest_w / src_w
        scaled_w = dest_w
        scaled_h = src_h * scale

    scaled_image = image.resize((int(scaled_w), int(scaled_h)), PIL.Image.ANTIALIAS)

    return scaled_image


def resize_crop_image(image, dest_w, dest_h, pad_when_tall=False):
    """
    :param image: PIL.Image
    :param dest_w: Target width
    :param dest_h: Target height
    :return: Scaled and cropped image
    """

    # Now we need to resize it
    dest_w = float(dest_w)
    dest_h = float(dest_h)
    dest_ratio = dest_w / dest_h

    # Calculate the apect ratio of the image
    src_w = float(image.size[0])
    src_h = float(image.size[1])
    src_ratio = src_w / src_h

    if src_ratio < dest_ratio:
        # Image is tall and thin - we need to scale to the right width and then crop
        scale = dest_w / src_w
        scaled_w = dest_w
        scaled_h = src_h * scale

        # Cropping values
        left = 0
        right = dest_w
        top = (scaled_h - dest_h) / 2.0
        bottom = top + dest_h
    else:
        # Image is short and wide - we need to scale to the right height and then crop
        scale = dest_h / src_h
        scaled_h = dest_h
        scaled_w = src_w * scale

        # Cropping values
        left = (scaled_w - dest_w) / 2.0
        right = left + dest_w
        top = 0
        bottom = dest_h

    if pad_when_tall:
        # Now, for images that are really tall and thin, we start to have issues as we only show a small section of them
        # (i.e. nasonex).  To deal with this we will resize and pad in this situation
        if (bottom - top) < (scaled_h * 0.66):
            log.info('Image would crop too much - returning padded image instead')
            return resize_pad_image(image, dest_w, dest_h)

    if src_w > dest_w or src_h > dest_h:
        # This means we are shrinking the image which is ok!
        scaled_image = image.resize((int(scaled_w), int(scaled_h)), PIL.Image.ANTIALIAS)
        cropped_image = scaled_image.crop((int(left), int(top), int(right), int(bottom)))

        return cropped_image

    elif scaled_w < src_w or scaled_h < src_h:
        # Just crop is as we don't want to stretch the image
        cropped_image = image.crop((int(left), int(top), int(right), int(bottom)))

        return cropped_image

    else:
        return image


def resize_pad_image(image, dest_w, dest_h):
    """
    Resize the image and pad to the correct aspect ratio.
    :param image: PIL.Image
    :param dest_w: Target width
    :param dest_h: Target height
    :return: Scaled and padded image
    """
    dest_w = float(dest_w)
    dest_h = float(dest_h)
    dest_ratio = dest_w / dest_h

    # Calculate the apect ratio of the image
    src_w = float(image.size[0])
    src_h = float(image.size[1])
    src_ratio = src_w / src_h

    if src_ratio < dest_ratio:
        # Image is tall and thin - we need to scale to the right height and then pad
        scale = dest_h / src_h
        scaled_h = dest_h
        scaled_w = src_w * scale

        offset = (int((dest_w - scaled_w) / 2), 0)

    else:
        # Image is short and wide - we need to scale to the right height and then crop
        scale = dest_w / src_w
        scaled_w = dest_w
        scaled_h = src_h * scale

        offset = (0, int((dest_h - scaled_h) / 2))

    scaled_image = image.resize((int(scaled_w), int(scaled_h)), PIL.Image.ANTIALIAS)
    # Normally we will want to copy the source mode for the destination image, but in some
    # cases the source image will use a Palletted (mode=='P') in which case we need to change
    # the mode
    mode = scaled_image.mode
    log.debug('Padding image with mode: "{}"'.format(mode))
    if mode == 'P':
        if 'transparency' in scaled_image.info:
            mode = 'RGBA'
        else:
            mode = 'RGB'

        scaled_image = scaled_image.convert(mode)
        log.debug('Changed mode from "P" to "{}"'.format(mode))

    # Get the pixel colour for coordinate (0,0)
    pixels = scaled_image.load()
    pad_colour = pixels[0, 0]
    
    padded_image = PIL.Image.new(mode, (int(dest_w), int(dest_h)), pad_colour)
    padded_image.paste(scaled_image, offset)

    return padded_image


def resize_image_to_fit_width(image, dest_w):
    """
    Resize and image to fit the passed in width, keeping the aspect ratio the same

    :param image: PIL.Image
    :param dest_w: The desired width
    """
    scale_factor = dest_w / image.size[0]
    dest_h = image.size[1] * scale_factor
    
    scaled_image = image.resize((int(dest_w), int(dest_h)), PIL.Image.ANTIALIAS)

    return scaled_image


def resize_image_to_fit_height(image, dest_h):
    """
    Resize and image to fit the passed in height, keeping the aspect ratio the same

    :param image: PIL.Image
    :param dest_h: The desired height
    """
    scale_factor = dest_h / image.size[1]
    dest_w = image.size[0] * scale_factor
    
    scaled_image = image.resize((int(dest_w), int(dest_h)), PIL.Image.ANTIALIAS)

    return scaled_image




"""
For doing geo ip lookups
"""

import logging

import geoip2.database
import IPy

__author__ = 'Yu Lee Paul (Little Fish Solutions LTD)'


log = logging.getLogger(__name__)

_reader = None


def init(geolite2_path):
    global _reader
    log.info('Loading geoip data from %s' % geolite2_path)
    _reader = geoip2.database.Reader(geolite2_path)


def get_geoip_data(ip):
    if should_ignore_ip_address(ip):
        return None
    return _reader.city(ip)


def should_ignore_ip_address(ip_address):
    ip = IPy.IP(ip_address)
    ip_type = ip.iptype()
    return ip_type == 'PRIVATE' or ip_type == 'LOOPBACK'



"""
Package containing general purpose utility code - basically the stuff that is used in every project

---

Library URL: git@lemon.com:pylib/lfs
"""

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

from .pager import Pager




"""
Class to prevent brute force attacks on the login
"""

import logging
import datetime
import threading

from .background import systemevent

__author__ = 'Yu Lee Paul (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class AttackProtect(object):
    def __init__(self, max_attempts, decrease_every, lock_duration):
        """
        :param max_attempts: maximum number of attempts allowed
        :param decrease_every: time (seconds) interval between decreasing attempts
        :param lock_duration: time (seconds) keys are locked for
        :return:
        """
        self.max_attempts = max_attempts
        self.decrease_every = decrease_every
        self.lock_duration = lock_duration

        # Dictionary maps keys onto number of logged attempts
        self.attempts = {}
        # Dictionary maps keys onto lock expiry times
        self.locks = {}
        # Lock object to ensure we don't read and write to this at the same time
        self.lock = threading.Lock()

        # Add system event to decrement counts and expire locks
        systemevent.queue_event(_AttackProtectServiceEvent(self))

    def log_attempt(self, key):
        """
        Log an attempt against key, incrementing the number of attempts for that key and potentially adding a lock to
        the lock table
        """
        with self.lock:
            if key not in self.attempts:
                self.attempts[key] = 1
            else:
                self.attempts[key] += 1

                if self.attempts[key] >= self.max_attempts:
                    log.info('Account %s locked due to too many login attempts' % key)
                    # lock account
                    self.locks[key] = datetime.datetime.utcnow() + datetime.timedelta(seconds=self.lock_duration)

    def is_allowed(self, key):
        """
        :return: True if the key is NOT locked, or false if it is locked
        """
        with self.lock:
            return key not in self.locks

    def service(self):
        """
        Decrease the countdowns, and remove any expired locks.  Should be called once every <decrease_every> seconds.
        """
        with self.lock:
            # Decrement / remove all attempts
            for key in list(self.attempts.keys()):
                log.debug('Decrementing count for %s' % key)
                if key in self.attempts:
                    if self.attempts[key] <= 1:
                        del self.attempts[key]
                    else:
                        self.attempts[key] -= 1

            # Remove expired locks
            now = datetime.datetime.utcnow()
            for key in list(self.locks.keys()):
                if key in self.locks and self.locks[key] < now:
                    log.info('Expiring login lock for %s' % key)
                    del self.locks[key]


class _AttackProtectServiceEvent(systemevent.SystemEvent):
    def __init__(self, attack_protect):
        self.attack_protect = attack_protect
        super(_AttackProtectServiceEvent, self).__init__(run_every=attack_protect.decrease_every)

    def process(self):
        self.attack_protect.service()



"""
A simple class to help with paging result sets
"""

import logging

from flask import request, url_for, Markup

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


class Pager(object):
    """
    Standard Pager used on back end of website.

    When viewing page 234 of 1000, the following page links will be displayed:
    
    1, 134, 184, 232, 233, 234, 235, 236, 284, 334, 1000
    """

    def __init__(self, page_size, page_number, query):
        self.page_size = page_size

        try:
            self.page_number = int(page_number)
        except ValueError:
            self.page_number = 1
        
        if self.page_number < 1:
            self.page_number = 1
        
        self.query = query
        
        # Do the paging here
        self.total_items = query.count()
        self.total_pages = (self.total_items - 1) // page_size + 1
        
        if self.page_number > self.total_pages:
            self.page_number = self.total_pages
        
        self.offset = self.page_size * (self.page_number - 1)
        if self.offset < 0:
            self.offset = 1

        self.items = query[self.offset:self.offset + self.page_size]
    
    def get_first_item_from_next_page(self):
        if self.has_next:
            return self.query[self.offset + self.page_size]

        return None
    
    def get_last_item_from_previous_page(self):
        if self.has_prev:
            return self.query[self.offset - 1]

        return None

    @property
    def has_prev(self):
        return self.page_number > 1
    
    @property
    def has_next(self):
        return self.page_number < self.total_pages
    
    @property
    def prev(self):
        return self.page_number - 1
    
    @property
    def next(self):
        return self.page_number + 1
    
    @property
    def page_link_numbers(self):
        pages = [1]
        
        if self.total_pages <= 1:
            return pages
        
        if self.page_number > 103:
            pages.append(self.page_number - 100)
        
        if self.page_number > 53:
            pages.append(self.page_number - 50)
        
        if self.page_number > 3:
            pages.append(self.page_number - 2)
        
        if self.page_number > 2:
            pages.append(self.page_number - 1)
            
        if self.page_number != 1 and self.page_number != self.total_pages:
            pages.append(self.page_number)
            
        if self.page_number < self.total_pages - 1:
            pages.append(self.page_number + 1)

        if self.page_number < self.total_pages - 2:
            pages.append(self.page_number + 2)
        
        if self.page_number < self.total_pages - 52:
            pages.append(self.page_number + 50)

        if self.page_number < self.total_pages - 102:
            pages.append(self.page_number + 100)
        
        pages.append(self.total_pages)
        return pages

    @property
    def empty(self):
        return self.total_pages == 0
    
    def get_full_page_url(self, page_number, scheme=None):
        """Get the full, external URL for this page, optinally with the passed in URL scheme"""
        args = dict(
            request.view_args,
            _external=True,
        )

        if scheme is not None:
            args['_scheme'] = scheme
        
        if page_number != 1:
            args['page'] = page_number

        return url_for(request.endpoint, **args)

    def get_canonical_url(self, scheme=None):
        """Get the canonical page URL"""
        return self.get_full_page_url(self.page_number, scheme=scheme)

    def render_prev_next_links(self, scheme=None):
        """Render the rel=prev and rel=next links to a Markup object for injection into a template"""
        output = ''

        if self.has_prev:
            output += '<link rel="prev" href="{}" />\n'.format(self.get_full_page_url(self.prev, scheme=scheme))
        
        if self.has_next:
            output += '<link rel="next" href="{}" />\n'.format(self.get_full_page_url(self.next, scheme=scheme))

        return Markup(output)

    def render_canonical_link(self, scheme=None):
        """Render the rel=canonical link to a Markup object for injection into a template"""
        return Markup('<link rel="canonical" href="{}" />'.format(self.get_canonical_url(scheme=scheme)))

    def render_seo_links(self, scheme=None):
        """Render the rel=canonical, rel=prev and rel=next links to a Markup object for injection into a template"""
        out = self.render_prev_next_links(scheme=scheme)

        if self.total_pages == 1:
            out += self.render_canonical_link(scheme=scheme)

        return out


class SimplePager(Pager):
    """
    Uses the same api as above, but displays a range of continuous page numbers.
    If you are on page 6 of 10 the following page numbers will be displayed:

    1, 2, 3, 4, 5, 6, 7, 8, 9, 10
    """

    def __init__(self, page_size, page_number, query, max_pages=12):
        """
        :param max_pages: The maximum number of page links to display
        """
        super().__init__(page_size, page_number, query)

        self.max_pages = max_pages
    
    @property
    def page_link_numbers(self):
        start = self.page_number - self.max_pages // 2 + 1
        if start < 1:
            start = 1

        end = start + self.max_pages - 1
        if end > self.total_pages:
            end = self.total_pages

            if start > 1:
                start = end - self.max_pages + 1

        return range(start, end + 1)


class ViewAllPager(object):
    """
    Uses the same API as pager, but lists all items on a single page.  This is to allow
    easy implementation of a "view all" function on a listing page
    """
    def __init__(self, query):
        self.page_number = 1

        self.query = query

        # Do the paging here
        self.total_items = query.count()
        self.page_size = self.total_items
        self.total_pages = 1

        self.offset = 0

        self.items = query.all()

    @property
    def has_prev(self):
        return False

    @property
    def has_next(self):
        return False

    @property
    def prev(self):
        return self.page_number - 1

    @property
    def next(self):
        return self.page_number + 1

    @property
    def page_link_numbers(self):
        return [1]




"""
This contains decorators and code to handle pages which require ssl
"""

import logging

from flask import request, redirect, current_app

__author__ = 'Stephen Brown (Little Fish Solutions LTD)'

log = logging.getLogger(__name__)


def ssl_required(fn):
    """
    Decorator - marks a route as needing ssl.

    NOTE: This most go BEFORE the route!
    """
    fn.ssl_required = True
    return fn


def ssl_allowed(fn):
    """
    Decorator - marks a route as allowing ssl, but not requiring it.  It can be served over http and https.

    NOTE: This must go BEFORE the route!
    """
    fn.ssl_allowed = True
    return fn


def ssl_disabled(fn):
    """
    Decorator - marks a route as not allowing ssl.

    NOTE: This must go BEFORE the route!
    """
    fn.ssl_disabled = True
    return fn


def handle_ssl_redirect():
    """
    Check if a route needs ssl, and redirect it if not.  Also redirects back to http for non-ssl routes.  Static routes
    are served as both http and https

    :return: A response to be returned or None
    """
    if request.endpoint and request.endpoint not in ['static', 'filemanager.static']:
        needs_ssl = False
        ssl_enabled = False
        view_function = current_app.view_functions[request.endpoint]
        if request.endpoint.startswith('admin.') or \
                (hasattr(view_function, 'ssl_required') and view_function.ssl_required):
            needs_ssl = True
            ssl_enabled = True

        if hasattr(view_function, 'ssl_allowed') and view_function.ssl_allowed:
            ssl_enabled = True

        if (hasattr(view_function, 'ssl_disabled') and view_function.ssl_disabled):
            needs_ssl = False
            ssl_enabled = False

        if current_app.config['SSL_ENABLED']:
            if needs_ssl and not request.is_secure:
                log.debug('Redirecting to https: %s' % request.endpoint)
                return redirect(request.url.replace("http://", "https://"))
            elif not ssl_enabled and request.is_secure:
                log.debug('Redirecting to http: %s' % request.endpoint)
                return redirect(request.url.replace("https://", "http://"))
        elif needs_ssl:
            log.info('Not redirecting to HTTPS for endpoint %s as SSL_ENABLED is set to False' % request.endpoint)
